
# Autogenerated by mlir-tblgen; don't manually edit.

from ._ods_common import _cext as _ods_cext
from ._ods_common import (
    equally_sized_accessor as _ods_equally_sized_accessor,
    get_default_loc_context as _ods_get_default_loc_context,
    get_op_results_or_values as _get_op_results_or_values,
    segmented_accessor as _ods_segmented_accessor,
)
_ods_ir = _ods_cext.ir
_ods_cext.globals.register_traceback_file_exclusion(__file__)

import builtins
from typing import Any as _Any, Sequence as _Sequence, Union as _Union, Optional as _Optional
import sys as _sys
if _sys.version_info >= (3, 12):
  from collections.abc import Buffer as _Buffer  # pytype: disable=not-supported-yet
else:
  try:
    from typing_extensions import Buffer as _Buffer
  except ImportError:
    _Buffer = _Any


@_ods_cext.register_dialect
class _Dialect(_ods_ir.Dialect):
  DIALECT_NAMESPACE = "async"

@_ods_cext.register_operation(_Dialect)
class AddToGroupOp(_ods_ir.OpView):
  r"""
  The `async.add_to_group` adds an async token or value to the async group.
  Returns the rank of the added element in the group. This rank is fixed
  for the group lifetime.
  
  Example:
  
  ```mlir
  %0 = async.create_group %size : !async.group
  %1 = ... : !async.token
  %2 = async.add_to_group %1, %0 : !async.token
  ```
  """

  OPERATION_NAME = "async.add_to_group"

  _ODS_REGIONS = (0, True)

  def __init__(self, operand: _ods_ir.Value, group: _ods_ir.Value, *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(operand)
    operands.append(group)
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def operand(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def group(self) -> _ods_ir.Value:
    return self.operation.operands[1]

  @builtins.property
  def rank(self) -> _ods_ir.OpResult[_ods_ir.IndexType]:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(AddToGroupOp)
class AddToGroupOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "async.add_to_group"

  @builtins.property
  def operand(self) -> _ods_ir.Value:
    return self.operands[0]

  @builtins.property
  def group(self) -> _ods_ir.Value:
    return self.operands[1]

def add_to_group(operand: _ods_ir.Value, group: _ods_ir.Value, *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult[_ods_ir.IndexType]:
  return AddToGroupOp(operand=operand, group=group, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class AwaitAllOp(_ods_ir.OpView):
  r"""
  The `async.await_all` operation waits until all the tokens or values in the
  group become ready.
  
  Example:
  
  ```mlir
  %0 = async.create_group %size : !async.group
  
  %1 = ... : !async.token
  %2 = async.add_to_group %1, %0 : !async.token
  
  %3 = ... : !async.token
  %4 = async.add_to_group %2, %0 : !async.token
  
  async.await_all %0
  ```
  """

  OPERATION_NAME = "async.await_all"

  _ODS_REGIONS = (0, True)

  def __init__(self, operand: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(operand)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def operand(self) -> _ods_ir.Value:
    return self.operation.operands[0]

@_ods_cext.register_op_adaptor(AwaitAllOp)
class AwaitAllOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "async.await_all"

  @builtins.property
  def operand(self) -> _ods_ir.Value:
    return self.operands[0]

def await_all(operand: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> AwaitAllOp:
  return AwaitAllOp(operand=operand, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class AwaitOp(_ods_ir.OpView):
  r"""
  The `async.await` operation waits until the argument becomes ready, and for
  the `async.value` arguments it unwraps the underlying value
  
  Example:
  
  ```mlir
  %0 = ... : !async.token
  async.await %0 : !async.token
  
  %1 = ... : !async.value<f32>
  %2 = async.await %1 : !async.value<f32>
  ```
  """

  OPERATION_NAME = "async.await"

  _ODS_REGIONS = (0, True)

  def __init__(self, result: _Optional[_ods_ir.Type], operand: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(operand)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    if result is not None: results.append(result)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def operand(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def result(self) -> _Optional[_ods_ir.OpResult]:
    return None if len(self.operation.results) < 1 else self.operation.results[0]

@_ods_cext.register_op_adaptor(AwaitOp)
class AwaitOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "async.await"

  @builtins.property
  def operand(self) -> _ods_ir.Value:
    return self.operands[0]

def await_(result: _Optional[_ods_ir.Type], operand: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _Union[_ods_ir.OpResult, _ods_ir.OpResultList, AwaitOp]:
  op = AwaitOp(result=result, operand=operand, loc=loc, ip=ip); results = op.results
  return results if len(results) > 1 else (results[0] if len(results) == 1 else op)

@_ods_cext.register_operation(_Dialect)
class CallOp(_ods_ir.OpView):
  r"""
  The `async.call` operation represents a direct call to an async function
  that is within the same symbol scope as the call. The operands and result
  types of the call must match the specified async function type. The callee
  is encoded as a symbol reference attribute named "callee".
  
  Example:
  
  ```mlir
  %2 = async.call @my_add(%0, %1) : (f32, f32) -> !async.value<f32>
  ```
  """

  OPERATION_NAME = "async.call"

  _ODS_REGIONS = (0, True)

  def __init__(self, result: _Sequence[_ods_ir.Type], callee: _Union[str, _ods_ir.FlatSymbolRefAttr], operands_: _Sequence[_ods_ir.Value], *, arg_attrs: _Optional[_Union[_Any, _ods_ir.ArrayAttr]] = None, res_attrs: _Optional[_Union[_Any, _ods_ir.ArrayAttr]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.extend(_get_op_results_or_values(operands_))
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["callee"] = (callee if (
    isinstance(callee, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('FlatSymbolRefAttr')) else
      _ods_ir.AttrBuilder.get('FlatSymbolRefAttr')(callee, context=_ods_context))
    if arg_attrs is not None: attributes["arg_attrs"] = (arg_attrs if (
        isinstance(arg_attrs, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('DictArrayAttr')) else
          _ods_ir.AttrBuilder.get('DictArrayAttr')(arg_attrs, context=_ods_context))
    if res_attrs is not None: attributes["res_attrs"] = (res_attrs if (
        isinstance(res_attrs, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('DictArrayAttr')) else
          _ods_ir.AttrBuilder.get('DictArrayAttr')(res_attrs, context=_ods_context))
    results = []
    results.extend(result)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def operands_(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operation.operands) - 1 + 1
    return self.operation.operands[0:0 + _ods_variadic_group_length]

  @builtins.property
  def callee(self) -> _ods_ir.FlatSymbolRefAttr:
    return self.operation.attributes["callee"]

  @callee.setter
  def callee(self, value: _ods_ir.FlatSymbolRefAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["callee"] = value

  @builtins.property
  def arg_attrs(self) -> _Optional[_ods_ir.ArrayAttr]:
    if "arg_attrs" not in self.operation.attributes:
      return None
    return self.operation.attributes["arg_attrs"]

  @arg_attrs.setter
  def arg_attrs(self, value: _Optional[_ods_ir.ArrayAttr]):
    if value is not None:
      self.operation.attributes["arg_attrs"] = value
    elif "arg_attrs" in self.operation.attributes:
      del self.operation.attributes["arg_attrs"]

  @arg_attrs.deleter
  def arg_attrs(self):
    del self.operation.attributes["arg_attrs"]

  @builtins.property
  def res_attrs(self) -> _Optional[_ods_ir.ArrayAttr]:
    if "res_attrs" not in self.operation.attributes:
      return None
    return self.operation.attributes["res_attrs"]

  @res_attrs.setter
  def res_attrs(self, value: _Optional[_ods_ir.ArrayAttr]):
    if value is not None:
      self.operation.attributes["res_attrs"] = value
    elif "res_attrs" in self.operation.attributes:
      del self.operation.attributes["res_attrs"]

  @res_attrs.deleter
  def res_attrs(self):
    del self.operation.attributes["res_attrs"]

@_ods_cext.register_op_adaptor(CallOp)
class CallOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "async.call"

  @builtins.property
  def operands_(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operands) - 1 + 1
    return self.operands[0:0 + _ods_variadic_group_length]

  @builtins.property
  def callee(self) -> _ods_ir.FlatSymbolRefAttr:
    return self.attributes["callee"]

  @builtins.property
  def arg_attrs(self) -> _Optional[_ods_ir.ArrayAttr]:
    if "arg_attrs" not in self.attributes:
      return None
    return self.attributes["arg_attrs"]

  @builtins.property
  def res_attrs(self) -> _Optional[_ods_ir.ArrayAttr]:
    if "res_attrs" not in self.attributes:
      return None
    return self.attributes["res_attrs"]

def call(result: _Sequence[_ods_ir.Type], callee: _Union[str, _ods_ir.FlatSymbolRefAttr], operands_: _Sequence[_ods_ir.Value], *, arg_attrs: _Optional[_Union[_Any, _ods_ir.ArrayAttr]] = None, res_attrs: _Optional[_Union[_Any, _ods_ir.ArrayAttr]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _Union[_ods_ir.OpResult, _ods_ir.OpResultList, CallOp]:
  op = CallOp(result=result, callee=callee, operands_=operands_, arg_attrs=arg_attrs, res_attrs=res_attrs, loc=loc, ip=ip); results = op.results
  return results if len(results) > 1 else (results[0] if len(results) == 1 else op)

@_ods_cext.register_operation(_Dialect)
class CoroBeginOp(_ods_ir.OpView):
  r"""
  The `async.coro.begin` allocates a coroutine frame and returns a handle to
  the coroutine.
  """

  OPERATION_NAME = "async.coro.begin"

  _ODS_REGIONS = (0, True)

  def __init__(self, id: _ods_ir.Value, *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(id)
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def id(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def handle(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(CoroBeginOp)
class CoroBeginOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "async.coro.begin"

  @builtins.property
  def id(self) -> _ods_ir.Value:
    return self.operands[0]

def coro_begin(id: _ods_ir.Value, *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return CoroBeginOp(id=id, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class CoroEndOp(_ods_ir.OpView):
  r"""
  The `async.coro.end` marks the point where a coroutine needs to return
  control back to the caller if it is not an initial invocation of the
  coroutine. It the start part of the coroutine is is no-op.
  """

  OPERATION_NAME = "async.coro.end"

  _ODS_REGIONS = (0, True)

  def __init__(self, handle: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(handle)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def handle(self) -> _ods_ir.Value:
    return self.operation.operands[0]

@_ods_cext.register_op_adaptor(CoroEndOp)
class CoroEndOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "async.coro.end"

  @builtins.property
  def handle(self) -> _ods_ir.Value:
    return self.operands[0]

def coro_end(handle: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> CoroEndOp:
  return CoroEndOp(handle=handle, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class CoroFreeOp(_ods_ir.OpView):
  r"""
  The `async.coro.free` deallocates the coroutine frame created by the
  async.coro.begin operation.
  """

  OPERATION_NAME = "async.coro.free"

  _ODS_REGIONS = (0, True)

  def __init__(self, id: _ods_ir.Value, handle: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(id)
    operands.append(handle)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def id(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def handle(self) -> _ods_ir.Value:
    return self.operation.operands[1]

@_ods_cext.register_op_adaptor(CoroFreeOp)
class CoroFreeOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "async.coro.free"

  @builtins.property
  def id(self) -> _ods_ir.Value:
    return self.operands[0]

  @builtins.property
  def handle(self) -> _ods_ir.Value:
    return self.operands[1]

def coro_free(id: _ods_ir.Value, handle: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> CoroFreeOp:
  return CoroFreeOp(id=id, handle=handle, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class CoroIdOp(_ods_ir.OpView):
  r"""
  The `async.coro.id` returns a switched-resume coroutine identifier.
  """

  OPERATION_NAME = "async.coro.id"

  _ODS_REGIONS = (0, True)

  def __init__(self, *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def id(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(CoroIdOp)
class CoroIdOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "async.coro.id"

def coro_id(*, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return CoroIdOp(results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class CoroSaveOp(_ods_ir.OpView):
  r"""
  The `async.coro.saves` saves the coroutine state.
  """

  OPERATION_NAME = "async.coro.save"

  _ODS_REGIONS = (0, True)

  def __init__(self, handle: _ods_ir.Value, *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(handle)
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def handle(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def state(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(CoroSaveOp)
class CoroSaveOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "async.coro.save"

  @builtins.property
  def handle(self) -> _ods_ir.Value:
    return self.operands[0]

def coro_save(handle: _ods_ir.Value, *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return CoroSaveOp(handle=handle, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class CoroSuspendOp(_ods_ir.OpView):
  r"""
  The `async.coro.suspend` suspends the coroutine and transfers control to the
  `suspend` successor. If suspended coroutine later resumed it will transfer
  control to the `resume` successor. If it is destroyed it will transfer
  control to the the `cleanup` successor.
  
  In switched-resume lowering coroutine can be already in resumed state when
  suspend operation is called, in this case control will be transferred to the
  `resume` successor skipping the `suspend` successor.
  """

  OPERATION_NAME = "async.coro.suspend"

  _ODS_REGIONS = (0, True)

  def __init__(self, state: _ods_ir.Value, suspendDest: _ods_ir.Block, resumeDest: _ods_ir.Block, cleanupDest: _ods_ir.Block, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(state)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = []
    _ods_successors.append(suspendDest)
    _ods_successors.append(resumeDest)
    _ods_successors.append(cleanupDest)
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def state(self) -> _ods_ir.Value:
    return self.operation.operands[0]

@_ods_cext.register_op_adaptor(CoroSuspendOp)
class CoroSuspendOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "async.coro.suspend"

  @builtins.property
  def state(self) -> _ods_ir.Value:
    return self.operands[0]

def coro_suspend(state: _ods_ir.Value, suspend_dest: _ods_ir.Block, resume_dest: _ods_ir.Block, cleanup_dest: _ods_ir.Block, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> CoroSuspendOp:
  return CoroSuspendOp(state=state, suspendDest=suspend_dest, resumeDest=resume_dest, cleanupDest=cleanup_dest, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class CreateGroupOp(_ods_ir.OpView):
  r"""
  The `async.create_group` allocates an empty async group. Async tokens or
  values can be added to this group later. The size of the group must be
  specified at construction time, and `await_all` operation will first
  wait until the number of added tokens or values reaches the group size.
  
  Example:
  
  ```mlir
  %size = ... : index
  %group = async.create_group %size : !async.group
  ...
  async.await_all %group
  ```
  """

  OPERATION_NAME = "async.create_group"

  _ODS_REGIONS = (0, True)

  def __init__(self, size: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(size)
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def size(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[0]

  @builtins.property
  def result(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(CreateGroupOp)
class CreateGroupOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "async.create_group"

  @builtins.property
  def size(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[0]

def create_group(size: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return CreateGroupOp(size=size, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class ExecuteOp(_ods_ir.OpView):
  r"""
   The `body` region attached to the `async.execute` operation semantically
   can be executed concurrently with the successor operation. In the followup
   example "compute0" can be executed concurrently with "compute1".
  
   The actual concurrency semantics depends on the dialect lowering to the
   executable format. Fully sequential execution ("compute0" completes before
   "compute1" starts) is a completely legal execution.
  
   Because concurrent execution is not guaranteed, it is illegal to create an
   implicit dependency from "compute1" to "compute0" (e.g. via shared global
   state). All dependencies must be made explicit with async execute arguments
   (`async.token` or `async.value`).
  
  `async.execute` operation takes `async.token` dependencies and `async.value`
   operands separately, and starts execution of the attached body region only
   when all tokens and values become ready.
  
   Example:
  
   ```mlir
   %dependency = ... : !async.token
   %value = ... : !async.value<f32>
  
   %token, %results =
     async.execute [%dependency](%value as %unwrapped: !async.value<f32>)
                -> !async.value<!some.type>
     {
       %0 = "compute0"(%unwrapped): (f32) -> !some.type
       async.yield %0 : !some.type
     }
  
   %1 = "compute1"(...) : !some.type
   ```
  
   In the example above asynchronous execution starts only after dependency
   token and value argument become ready. Unwrapped value passed to the
   attached body region as an %unwrapped value of f32 type.
  """

  OPERATION_NAME = "async.execute"

  _ODS_OPERAND_SEGMENTS = [-1,-1,]

  _ODS_REGIONS = (1, True)

  def __init__(self, token: _ods_ir.Type, bodyResults: _Sequence[_ods_ir.Type], dependencies: _Sequence[_ods_ir.Value], bodyOperands: _Sequence[_ods_ir.Value], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(_get_op_results_or_values(dependencies))
    operands.append(_get_op_results_or_values(bodyOperands))
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    results.append(token)
    results.extend(bodyResults)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def dependencies(self) -> _ods_ir.OpOperandList:
    operand_range = _ods_segmented_accessor(
         self.operation.operands,
         self.operation.attributes["operandSegmentSizes"], 0)
    return operand_range

  @builtins.property
  def bodyOperands(self) -> _ods_ir.OpOperandList:
    operand_range = _ods_segmented_accessor(
         self.operation.operands,
         self.operation.attributes["operandSegmentSizes"], 1)
    return operand_range

  @builtins.property
  def token(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

  @builtins.property
  def bodyResults(self) -> _ods_ir.OpResultList:
    _ods_variadic_group_length = len(self.operation.results) - 2 + 1
    return self.operation.results[1:1 + _ods_variadic_group_length]

  @builtins.property
  def bodyRegion(self) -> _ods_ir.Region:
    return self.regions[0]

@_ods_cext.register_op_adaptor(ExecuteOp)
class ExecuteOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "async.execute"

  @builtins.property
  def dependencies(self) -> _ods_ir.OpOperandList:
    operand_range = _ods_segmented_accessor(
         self.operands,
         self.attributes["operandSegmentSizes"], 0)
    return operand_range

  @builtins.property
  def bodyOperands(self) -> _ods_ir.OpOperandList:
    operand_range = _ods_segmented_accessor(
         self.operands,
         self.attributes["operandSegmentSizes"], 1)
    return operand_range

def execute(token: _ods_ir.Type, body_results: _Sequence[_ods_ir.Type], dependencies: _Sequence[_ods_ir.Value], body_operands: _Sequence[_ods_ir.Value], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _Union[_ods_ir.OpResult, _ods_ir.OpResultList, ExecuteOp]:
  op = ExecuteOp(token=token, bodyResults=body_results, dependencies=dependencies, bodyOperands=body_operands, loc=loc, ip=ip); results = op.results
  return results if len(results) > 1 else (results[0] if len(results) == 1 else op)

@_ods_cext.register_operation(_Dialect)
class FuncOp(_ods_ir.OpView):
  r"""
  An async function is like a normal function, but supports non-blocking
  await. Internally, async function is lowered to the LLVM coroutinue with
  async runtime intrinsic. It can return an async token and/or async values.
  The token represents the execution state of async function and can be used
  when users want to express dependencies on some side effects, e.g.,
  the token becomes available once every thing in the func body is executed.
  
  Example:
  
  ```mlir
  // Async function can't return void, it always must be some async thing.
  async.func @async.0() -> !async.token {
    return
  }
  
  // Function returns only async value.
  async.func @async.1() -> !async.value<i32> {
    %0 = arith.constant 42 : i32
    return %0 : i32
  }
  
  // Implicit token can be added to return types.
  async.func @async.2() -> !async.token, !async.value<i32> {
    %0 = arith.constant 42 : i32
    return %0 : i32
  }
  ```
  """

  OPERATION_NAME = "async.func"

  _ODS_REGIONS = (1, True)

  def __init__(self, sym_name: _Union[str, _ods_ir.StringAttr], function_type: _Union[_Any, _ods_ir.TypeAttr], *, sym_visibility: _Optional[_Union[str, _ods_ir.StringAttr]] = None, arg_attrs: _Optional[_Union[_Any, _ods_ir.ArrayAttr]] = None, res_attrs: _Optional[_Union[_Any, _ods_ir.ArrayAttr]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["sym_name"] = (sym_name if (
    isinstance(sym_name, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('SymbolNameAttr')) else
      _ods_ir.AttrBuilder.get('SymbolNameAttr')(sym_name, context=_ods_context))
    attributes["function_type"] = (function_type if (
    isinstance(function_type, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('anonymous_454')) else
      _ods_ir.AttrBuilder.get('anonymous_454')(function_type, context=_ods_context))
    if sym_visibility is not None: attributes["sym_visibility"] = (sym_visibility if (
        isinstance(sym_visibility, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('StrAttr')) else
          _ods_ir.AttrBuilder.get('StrAttr')(sym_visibility, context=_ods_context))
    if arg_attrs is not None: attributes["arg_attrs"] = (arg_attrs if (
        isinstance(arg_attrs, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('DictArrayAttr')) else
          _ods_ir.AttrBuilder.get('DictArrayAttr')(arg_attrs, context=_ods_context))
    if res_attrs is not None: attributes["res_attrs"] = (res_attrs if (
        isinstance(res_attrs, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('DictArrayAttr')) else
          _ods_ir.AttrBuilder.get('DictArrayAttr')(res_attrs, context=_ods_context))
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def sym_name(self) -> _ods_ir.StringAttr:
    return self.operation.attributes["sym_name"]

  @sym_name.setter
  def sym_name(self, value: _ods_ir.StringAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["sym_name"] = value

  @builtins.property
  def function_type(self) -> _ods_ir.TypeAttr:
    return self.operation.attributes["function_type"]

  @function_type.setter
  def function_type(self, value: _ods_ir.TypeAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["function_type"] = value

  @builtins.property
  def sym_visibility(self) -> _Optional[_ods_ir.StringAttr]:
    if "sym_visibility" not in self.operation.attributes:
      return None
    return self.operation.attributes["sym_visibility"]

  @sym_visibility.setter
  def sym_visibility(self, value: _Optional[_ods_ir.StringAttr]):
    if value is not None:
      self.operation.attributes["sym_visibility"] = value
    elif "sym_visibility" in self.operation.attributes:
      del self.operation.attributes["sym_visibility"]

  @sym_visibility.deleter
  def sym_visibility(self):
    del self.operation.attributes["sym_visibility"]

  @builtins.property
  def arg_attrs(self) -> _Optional[_ods_ir.ArrayAttr]:
    if "arg_attrs" not in self.operation.attributes:
      return None
    return self.operation.attributes["arg_attrs"]

  @arg_attrs.setter
  def arg_attrs(self, value: _Optional[_ods_ir.ArrayAttr]):
    if value is not None:
      self.operation.attributes["arg_attrs"] = value
    elif "arg_attrs" in self.operation.attributes:
      del self.operation.attributes["arg_attrs"]

  @arg_attrs.deleter
  def arg_attrs(self):
    del self.operation.attributes["arg_attrs"]

  @builtins.property
  def res_attrs(self) -> _Optional[_ods_ir.ArrayAttr]:
    if "res_attrs" not in self.operation.attributes:
      return None
    return self.operation.attributes["res_attrs"]

  @res_attrs.setter
  def res_attrs(self, value: _Optional[_ods_ir.ArrayAttr]):
    if value is not None:
      self.operation.attributes["res_attrs"] = value
    elif "res_attrs" in self.operation.attributes:
      del self.operation.attributes["res_attrs"]

  @res_attrs.deleter
  def res_attrs(self):
    del self.operation.attributes["res_attrs"]

  @builtins.property
  def body(self) -> _ods_ir.Region:
    return self.regions[0]

@_ods_cext.register_op_adaptor(FuncOp)
class FuncOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "async.func"

  @builtins.property
  def sym_name(self) -> _ods_ir.StringAttr:
    return self.attributes["sym_name"]

  @builtins.property
  def function_type(self) -> _ods_ir.TypeAttr:
    return self.attributes["function_type"]

  @builtins.property
  def sym_visibility(self) -> _Optional[_ods_ir.StringAttr]:
    if "sym_visibility" not in self.attributes:
      return None
    return self.attributes["sym_visibility"]

  @builtins.property
  def arg_attrs(self) -> _Optional[_ods_ir.ArrayAttr]:
    if "arg_attrs" not in self.attributes:
      return None
    return self.attributes["arg_attrs"]

  @builtins.property
  def res_attrs(self) -> _Optional[_ods_ir.ArrayAttr]:
    if "res_attrs" not in self.attributes:
      return None
    return self.attributes["res_attrs"]

def func(sym_name: _Union[str, _ods_ir.StringAttr], function_type: _Union[_Any, _ods_ir.TypeAttr], *, sym_visibility: _Optional[_Union[str, _ods_ir.StringAttr]] = None, arg_attrs: _Optional[_Union[_Any, _ods_ir.ArrayAttr]] = None, res_attrs: _Optional[_Union[_Any, _ods_ir.ArrayAttr]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> FuncOp:
  return FuncOp(sym_name=sym_name, function_type=function_type, sym_visibility=sym_visibility, arg_attrs=arg_attrs, res_attrs=res_attrs, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class ReturnOp(_ods_ir.OpView):
  r"""
  The `async.return` is a special terminator operation for Async function.
  
  Example:
  
  ```mlir
  async.func @foo() : !async.token {
    return
  }
  ```
  """

  OPERATION_NAME = "async.return"

  _ODS_REGIONS = (0, True)

  def __init__(self, operands_: _Sequence[_ods_ir.Value], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.extend(_get_op_results_or_values(operands_))
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def operands_(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operation.operands) - 1 + 1
    return self.operation.operands[0:0 + _ods_variadic_group_length]

@_ods_cext.register_op_adaptor(ReturnOp)
class ReturnOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "async.return"

  @builtins.property
  def operands_(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operands) - 1 + 1
    return self.operands[0:0 + _ods_variadic_group_length]

def return_(operands_: _Sequence[_ods_ir.Value], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> ReturnOp:
  return ReturnOp(operands_=operands_, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class RuntimeAddRefOp(_ods_ir.OpView):
  r"""
  The `async.runtime.add_ref` operation adds a reference(s) to async value
  (token, value or group).
  """

  OPERATION_NAME = "async.runtime.add_ref"

  _ODS_REGIONS = (0, True)

  def __init__(self, operand: _ods_ir.Value, count: _Union[int, _ods_ir.IntegerAttr], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(operand)
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["count"] = (count if (
    isinstance(count, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('I64Attr')) else
      _ods_ir.AttrBuilder.get('I64Attr')(count, context=_ods_context))
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def operand(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def count(self) -> _ods_ir.IntegerAttr:
    return self.operation.attributes["count"]

  @count.setter
  def count(self, value: _ods_ir.IntegerAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["count"] = value

@_ods_cext.register_op_adaptor(RuntimeAddRefOp)
class RuntimeAddRefOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "async.runtime.add_ref"

  @builtins.property
  def operand(self) -> _ods_ir.Value:
    return self.operands[0]

  @builtins.property
  def count(self) -> _ods_ir.IntegerAttr:
    return self.attributes["count"]

def runtime_add_ref(operand: _ods_ir.Value, count: _Union[int, _ods_ir.IntegerAttr], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> RuntimeAddRefOp:
  return RuntimeAddRefOp(operand=operand, count=count, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class RuntimeAddToGroupOp(_ods_ir.OpView):
  r"""
  The `async.runtime.add_to_group` adds an async token or value to the async
  group. Returns the rank of the added element in the group.
  """

  OPERATION_NAME = "async.runtime.add_to_group"

  _ODS_REGIONS = (0, True)

  def __init__(self, operand: _ods_ir.Value, group: _ods_ir.Value, *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(operand)
    operands.append(group)
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def operand(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def group(self) -> _ods_ir.Value:
    return self.operation.operands[1]

  @builtins.property
  def rank(self) -> _ods_ir.OpResult[_ods_ir.IndexType]:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(RuntimeAddToGroupOp)
class RuntimeAddToGroupOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "async.runtime.add_to_group"

  @builtins.property
  def operand(self) -> _ods_ir.Value:
    return self.operands[0]

  @builtins.property
  def group(self) -> _ods_ir.Value:
    return self.operands[1]

def runtime_add_to_group(operand: _ods_ir.Value, group: _ods_ir.Value, *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult[_ods_ir.IndexType]:
  return RuntimeAddToGroupOp(operand=operand, group=group, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class RuntimeAwaitAndResumeOp(_ods_ir.OpView):
  r"""
  The `async.runtime.await_and_resume` operation awaits for the operand to
  become available or error and resumes the coroutine on a thread managed by
  the runtime.
  """

  OPERATION_NAME = "async.runtime.await_and_resume"

  _ODS_REGIONS = (0, True)

  def __init__(self, operand: _ods_ir.Value, handle: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(operand)
    operands.append(handle)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def operand(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def handle(self) -> _ods_ir.Value:
    return self.operation.operands[1]

@_ods_cext.register_op_adaptor(RuntimeAwaitAndResumeOp)
class RuntimeAwaitAndResumeOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "async.runtime.await_and_resume"

  @builtins.property
  def operand(self) -> _ods_ir.Value:
    return self.operands[0]

  @builtins.property
  def handle(self) -> _ods_ir.Value:
    return self.operands[1]

def runtime_await_and_resume(operand: _ods_ir.Value, handle: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> RuntimeAwaitAndResumeOp:
  return RuntimeAwaitAndResumeOp(operand=operand, handle=handle, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class RuntimeAwaitOp(_ods_ir.OpView):
  r"""
  The `async.runtime.await` operation blocks the caller thread until the
  operand becomes available or error.
  """

  OPERATION_NAME = "async.runtime.await"

  _ODS_REGIONS = (0, True)

  def __init__(self, operand: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(operand)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def operand(self) -> _ods_ir.Value:
    return self.operation.operands[0]

@_ods_cext.register_op_adaptor(RuntimeAwaitOp)
class RuntimeAwaitOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "async.runtime.await"

  @builtins.property
  def operand(self) -> _ods_ir.Value:
    return self.operands[0]

def runtime_await(operand: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> RuntimeAwaitOp:
  return RuntimeAwaitOp(operand=operand, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class RuntimeCreateGroupOp(_ods_ir.OpView):
  r"""
  The `async.runtime.create_group` operation creates an async dialect group
  of the given size. Group created in the empty state.
  """

  OPERATION_NAME = "async.runtime.create_group"

  _ODS_REGIONS = (0, True)

  def __init__(self, size: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(size)
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def size(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operation.operands[0]

  @builtins.property
  def result(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(RuntimeCreateGroupOp)
class RuntimeCreateGroupOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "async.runtime.create_group"

  @builtins.property
  def size(self) -> _ods_ir.Value[_ods_ir.IndexType]:
    return self.operands[0]

def runtime_create_group(size: _ods_ir.Value[_ods_ir.IndexType], *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return RuntimeCreateGroupOp(size=size, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class RuntimeCreateOp(_ods_ir.OpView):
  r"""
  The `async.runtime.create` operation creates an async dialect token or
  value. Tokens and values are created in the non-ready state.
  """

  OPERATION_NAME = "async.runtime.create"

  _ODS_REGIONS = (0, True)

  def __init__(self, result: _ods_ir.Type, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    results.append(result)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def result(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(RuntimeCreateOp)
class RuntimeCreateOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "async.runtime.create"

def runtime_create(result: _ods_ir.Type, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return RuntimeCreateOp(result=result, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class RuntimeDropRefOp(_ods_ir.OpView):
  r"""
  The `async.runtime.drop_ref` operation drops a reference(s) to async value
  (token, value or group).
  """

  OPERATION_NAME = "async.runtime.drop_ref"

  _ODS_REGIONS = (0, True)

  def __init__(self, operand: _ods_ir.Value, count: _Union[int, _ods_ir.IntegerAttr], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(operand)
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["count"] = (count if (
    isinstance(count, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('I64Attr')) else
      _ods_ir.AttrBuilder.get('I64Attr')(count, context=_ods_context))
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def operand(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def count(self) -> _ods_ir.IntegerAttr:
    return self.operation.attributes["count"]

  @count.setter
  def count(self, value: _ods_ir.IntegerAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["count"] = value

@_ods_cext.register_op_adaptor(RuntimeDropRefOp)
class RuntimeDropRefOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "async.runtime.drop_ref"

  @builtins.property
  def operand(self) -> _ods_ir.Value:
    return self.operands[0]

  @builtins.property
  def count(self) -> _ods_ir.IntegerAttr:
    return self.attributes["count"]

def runtime_drop_ref(operand: _ods_ir.Value, count: _Union[int, _ods_ir.IntegerAttr], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> RuntimeDropRefOp:
  return RuntimeDropRefOp(operand=operand, count=count, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class RuntimeIsErrorOp(_ods_ir.OpView):
  r"""
  The `async.runtime.is_error` operation returns true if the token, value or
  group (any of the async runtime values) is in the error state. It is the
  caller responsibility to check error state after the call to `await` or
  resuming after `await_and_resume`.
  """

  OPERATION_NAME = "async.runtime.is_error"

  _ODS_REGIONS = (0, True)

  def __init__(self, operand: _ods_ir.Value, *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(operand)
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def operand(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def is_error(self) -> _ods_ir.OpResult[_ods_ir.IntegerType]:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(RuntimeIsErrorOp)
class RuntimeIsErrorOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "async.runtime.is_error"

  @builtins.property
  def operand(self) -> _ods_ir.Value:
    return self.operands[0]

def runtime_is_error(operand: _ods_ir.Value, *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult[_ods_ir.IntegerType]:
  return RuntimeIsErrorOp(operand=operand, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class RuntimeLoadOp(_ods_ir.OpView):
  r"""
  The `async.runtime.load` operation loads the value from the runtime
  async.value storage.
  """

  OPERATION_NAME = "async.runtime.load"

  _ODS_REGIONS = (0, True)

  def __init__(self, storage: _ods_ir.Value, *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(storage)
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def storage(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def result(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(RuntimeLoadOp)
class RuntimeLoadOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "async.runtime.load"

  @builtins.property
  def storage(self) -> _ods_ir.Value:
    return self.operands[0]

def runtime_load(storage: _ods_ir.Value, *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return RuntimeLoadOp(storage=storage, results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class RuntimeNumWorkerThreadsOp(_ods_ir.OpView):
  r"""
  The `async.runtime.num_worker_threads` operation gets the number of threads
  in the threadpool from the runtime.
  """

  OPERATION_NAME = "async.runtime.num_worker_threads"

  _ODS_REGIONS = (0, True)

  def __init__(self, *, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def result(self) -> _ods_ir.OpResult[_ods_ir.IndexType]:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(RuntimeNumWorkerThreadsOp)
class RuntimeNumWorkerThreadsOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "async.runtime.num_worker_threads"

def runtime_num_worker_threads(*, results: _Optional[_Sequence[_ods_ir.Type]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult[_ods_ir.IndexType]:
  return RuntimeNumWorkerThreadsOp(results=results, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class RuntimeResumeOp(_ods_ir.OpView):
  r"""
  The `async.runtime.resume` operation resumes the coroutine on a thread
  managed by the runtime.
  """

  OPERATION_NAME = "async.runtime.resume"

  _ODS_REGIONS = (0, True)

  def __init__(self, handle: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(handle)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def handle(self) -> _ods_ir.Value:
    return self.operation.operands[0]

@_ods_cext.register_op_adaptor(RuntimeResumeOp)
class RuntimeResumeOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "async.runtime.resume"

  @builtins.property
  def handle(self) -> _ods_ir.Value:
    return self.operands[0]

def runtime_resume(handle: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> RuntimeResumeOp:
  return RuntimeResumeOp(handle=handle, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class RuntimeSetAvailableOp(_ods_ir.OpView):
  r"""
  The `async.runtime.set_available` operation switches async token or value
  state to available.
  """

  OPERATION_NAME = "async.runtime.set_available"

  _ODS_REGIONS = (0, True)

  def __init__(self, operand: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(operand)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def operand(self) -> _ods_ir.Value:
    return self.operation.operands[0]

@_ods_cext.register_op_adaptor(RuntimeSetAvailableOp)
class RuntimeSetAvailableOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "async.runtime.set_available"

  @builtins.property
  def operand(self) -> _ods_ir.Value:
    return self.operands[0]

def runtime_set_available(operand: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> RuntimeSetAvailableOp:
  return RuntimeSetAvailableOp(operand=operand, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class RuntimeSetErrorOp(_ods_ir.OpView):
  r"""
  The `async.runtime.set_error` operation switches async token or value
  state to error.
  """

  OPERATION_NAME = "async.runtime.set_error"

  _ODS_REGIONS = (0, True)

  def __init__(self, operand: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(operand)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def operand(self) -> _ods_ir.Value:
    return self.operation.operands[0]

@_ods_cext.register_op_adaptor(RuntimeSetErrorOp)
class RuntimeSetErrorOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "async.runtime.set_error"

  @builtins.property
  def operand(self) -> _ods_ir.Value:
    return self.operands[0]

def runtime_set_error(operand: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> RuntimeSetErrorOp:
  return RuntimeSetErrorOp(operand=operand, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class RuntimeStoreOp(_ods_ir.OpView):
  r"""
  The `async.runtime.store` operation stores the value into the runtime
  async.value storage.
  """

  OPERATION_NAME = "async.runtime.store"

  _ODS_REGIONS = (0, True)

  def __init__(self, value: _ods_ir.Value, storage: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(value)
    operands.append(storage)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def value(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def storage(self) -> _ods_ir.Value:
    return self.operation.operands[1]

@_ods_cext.register_op_adaptor(RuntimeStoreOp)
class RuntimeStoreOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "async.runtime.store"

  @builtins.property
  def value(self) -> _ods_ir.Value:
    return self.operands[0]

  @builtins.property
  def storage(self) -> _ods_ir.Value:
    return self.operands[1]

def runtime_store(value: _ods_ir.Value, storage: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> RuntimeStoreOp:
  return RuntimeStoreOp(value=value, storage=storage, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class YieldOp(_ods_ir.OpView):
  r"""
  The `async.yield` is a special terminator operation for the block inside
  `async.execute` operation.
  """

  OPERATION_NAME = "async.yield"

  _ODS_REGIONS = (0, True)

  def __init__(self, operands_: _Sequence[_ods_ir.Value], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.extend(_get_op_results_or_values(operands_))
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def operands_(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operation.operands) - 1 + 1
    return self.operation.operands[0:0 + _ods_variadic_group_length]

@_ods_cext.register_op_adaptor(YieldOp)
class YieldOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "async.yield"

  @builtins.property
  def operands_(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operands) - 1 + 1
    return self.operands[0:0 + _ods_variadic_group_length]

def yield_(operands_: _Sequence[_ods_ir.Value], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> YieldOp:
  return YieldOp(operands_=operands_, loc=loc, ip=ip)



#  Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
#  See https://llvm.org/LICENSE.txt for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

import inspect
from functools import wraps

from ..dialects._ods_common import get_op_result_or_op_results
from ..ir import Type, InsertionPoint


def op_region_builder(op, op_region, terminator=None):
    def builder_wrapper(body_builder):
        # Add a block with block args having types determined by type hints on the wrapped function.
        if len(op_region.blocks) == 0:
            sig = inspect.signature(body_builder)
            types = [p.annotation for p in sig.parameters.values()]
            if not (
                len(types) == len(sig.parameters)
                and all(isinstance(t, Type) for t in types)
            ):
                raise ValueError(
                    f"for {body_builder=} either missing a type annotation or type annotation isn't a mlir type: {sig}"
                )

            op_region.blocks.append(*types)

        with InsertionPoint(op_region.blocks[0]):
            results = body_builder(*list(op_region.blocks[0].arguments))

        with InsertionPoint(list(op_region.blocks)[-1]):
            if terminator is not None:
                res = []
                if isinstance(results, (tuple, list)):
                    res.extend(results)
                elif results is not None:
                    res.append(results)
                terminator(res)

        return get_op_result_or_op_results(op)

    return builder_wrapper


def region_op(op_constructor, terminator=None):
    """Decorator to define an MLIR Op specified as a python function.

    Requires that an `mlir.ir.InsertionPoint` and `mlir.ir.Location` are
    active for the current thread (i.e. established in a `with` block).

    Supports "naked" usage i.e., no parens if no args need to be passed to the Op constructor.

    When applied as a decorator to a Python function, an entry block will
    be constructed for the Op with types as specified **as type hints on the args of the function**.
    The block arguments will be passed positionally to the Python function.

    If a terminator is specified then the return from the decorated function will be passed
    to the terminator as the last statement in the entry block. Note, the API for the terminator
    is a (possibly empty) list; terminator accepting single values should be wrapped in a
    `lambda args: term(args[0])`

    The identifier (name) of the function will become:
    1. A single value result if the Op returns a single value;
    2. An OpResultList (as a list) if the Op returns multiple values;
    3. The Operation if the Op returns no results.

    See examples in tensor.py and transform.extras.
    """

    def op_decorator(*args, **kwargs):
        op = op_constructor(*args, **kwargs)
        op_region = op.regions[0]

        return op_region_builder(op, op_region, terminator)

    @wraps(op_decorator)
    def maybe_no_args(*args, **kwargs):
        if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
            return op_decorator()(args[0])
        else:
            return op_decorator(*args, **kwargs)

    return maybe_no_args



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

# DEEPBIND-preload bundled LLVM/MLIR libs before any MLIR binding is
# imported, so their symbols cannot be preempted by another in-process
# LLVM. See #170.
_mlir_deepbind_handles = []


def _preload_mlir_libs_with_deepbind():
    import sys

    if not sys.platform.startswith("linux"):
        return  # RTLD_DEEPBIND is Linux-only

    import ctypes
    import os
    from pathlib import Path

    mode = (
        os.RTLD_NOW | os.RTLD_LOCAL | getattr(os, "RTLD_DEEPBIND", 0)  # missing on musl / uclibc
    )
    libs_dir = Path(__file__).parent / "_mlir" / "_mlir_libs"

    # CAPI first: it has no sibling DT_NEEDED so DEEPBIND applies at
    # first load; the other libs DT_NEEDED it and reuse the handle.
    for name in (
        "libMLIRPythonCAPI.so",
        "libMLIRPythonSupport-numba_cuda_mlir.so",
        "libMLIRToLLVM70.so",
        "libMLIRModernToNVVM.so",
    ):
        lib = libs_dir / name
        if lib.exists():
            _mlir_deepbind_handles.append(ctypes.CDLL(str(lib), mode=mode))


_preload_mlir_libs_with_deepbind()
del _preload_mlir_libs_with_deepbind

from numba_cuda_mlir._version import __version__
from numba_cuda_mlir.mlir import make_nanobind_metaclass_inheritable
from numba_cuda_mlir.numba_cuda.np.numpy_support import carray, farray  # noqa: F401

make_nanobind_metaclass_inheritable()

__all__ = ["cuda", "carray", "farray", "__version__"]



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Caching support for numba_cuda_mlir.

This module provides cache implementations for MLIRDispatcher, allowing
compiled CUDA kernels to be cached to disk and reloaded in subsequent runs.
"""

from numba_cuda_mlir.numba_cuda import typing
from numba_cuda_mlir.numba_cuda.core.caching import Cache, CacheImpl, NullCache


class CachedCompileResult:
    """
    A minimal compile result reconstructed from cached data.

    This provides just enough interface to be used by MLIRDispatcher
    without requiring the full compilation infrastructure.
    """

    def __init__(self, signature, metadata):
        self.signature = signature
        self.metadata = metadata
        # Provide a codegen attribute for cache key computation
        self.codegen = _CachedCodegen()
        # entry_point is needed for recompile() to work
        self.entry_point = metadata.get("func_name")


class _CachedCodegen:
    """Minimal codegen for cache key computation."""

    def magic_tuple(self):
        # Return a consistent tuple for MLIR-compiled code
        from numba_cuda_mlir.tools import (
            get_gpu_compute_capability,
            get_cuda_runtime_version,
        )

        cc = get_gpu_compute_capability(tuple)
        return (get_cuda_runtime_version(), cc)


class MLIRCacheImpl(CacheImpl):
    """
    Cache implementation for MLIR-compiled CUDA kernels.

    Handles serialization and deserialization of compile results.
    """

    def reduce(self, cres):
        """Serialize a compile result for caching."""
        return {
            "signature_args": cres.signature.args,
            "signature_return_type": cres.signature.return_type,
            "cubin": cres.metadata.get("cubin"),
            "ptx": cres.metadata.get("ptx"),
            "ltoir": cres.metadata.get("ltoir"),
            "func_name": cres.metadata.get("func_name"),
            "mlir_module_optimized": cres.metadata.get("mlir_module_optimized"),
            "needs_nrt": cres.metadata.get("needs_nrt"),
            "nrt_inline": cres.metadata.get("nrt_inline"),
            "targetoptions": cres.metadata.get("targetoptions", {}),
            "gpu_target": cres.metadata.get("gpu_target"),
        }

    def rebuild(self, target_context, payload):
        """Deserialize a compile result from cache."""
        signature_args = payload["signature_args"]
        signature_return_type = payload["signature_return_type"]
        cubin = payload["cubin"]
        ptx = payload["ptx"]
        ltoir = payload.get("ltoir")
        func_name = payload["func_name"]
        mlir_module_optimized = payload.get("mlir_module_optimized")
        needs_nrt = payload.get("needs_nrt")
        nrt_inline = payload.get("nrt_inline")
        targetoptions = payload.get("targetoptions", {})
        gpu_target = payload.get("gpu_target")

        signature = typing.signature(signature_return_type, *signature_args)

        return CachedCompileResult(
            signature=signature,
            metadata={
                "cubin": cubin,
                "ptx": ptx,
                "ltoir": ltoir,
                "func_name": func_name,
                "mlir_module_optimized": mlir_module_optimized,
                "needs_nrt": needs_nrt,
                "nrt_inline": nrt_inline,
                "targetoptions": targetoptions,
                "gpu_target": gpu_target,
            },
        )

    def check_cachable(self, cres):
        """Check if a compile result can be cached."""
        targetoptions = cres.metadata.get("targetoptions", {})
        link = targetoptions.get("link", [])
        if link:
            raise RuntimeError("Cannot pickle CUDACodeLibrary with linking files")
        return True


class MLIRCache(Cache):
    """
    Cache for MLIR-compiled CUDA kernels.

    Uses the standard numba caching infrastructure with MLIR-specific
    serialization.
    """

    _impl_class = MLIRCacheImpl

    def __init__(self, py_func, targetoptions=None):
        self._targetoptions = targetoptions if targetoptions is not None else {}
        super().__init__(py_func)

    def _index_key(self, sig, codegen):
        key = super()._index_key(sig, codegen)
        targetoptions = self._targetoptions
        from numba_cuda_mlir.tools import resolve_gpu_target

        gpu_target = resolve_gpu_target(targetoptions)
        option_key = (
            ("lto_explicit", targetoptions.get("_lto_explicit", False)),
            ("lto", targetoptions.get("lto")),
            ("chip", gpu_target["chip"]),
            ("launch_bounds", targetoptions.get("launch_bounds")),
        )
        return (*key, option_key)


# Re-export NullCache for convenience
__all__ = ["MLIRCache", "MLIRCacheImpl", "NullCache", "CachedCompileResult"]



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""
Lowering for NumPy Record (structured dtype) types.

Records are represented as pointers to byte arrays. Field access is done via
byte offset calculations using LLVM GEP and bitcast operations.
"""

from numba_cuda_mlir.numba_cuda import types
from numba_cuda_mlir.types import Record, NestedArray
from numba_cuda_mlir.numba_cuda.core import ir as numba_ir
import operator

from numba_cuda_mlir._mlir.dialects import (
    llvm,
    arith,
    memref as memref_dialect,
    scf,
)
from numba_cuda_mlir._mlir import ir
from numba_cuda_mlir._mlir.extras import types as T
from numba_cuda_mlir._mlir.dialects import arith as arith_dialect
from numba_cuda_mlir.lowering_utilities import storage_itemsize_bytes


# For llvm.getelementptr - dynamic offset marker
LLVM_DYNAMIC = -2147483648

from numba_cuda_mlir.lowering_registry import LoweringRegistry
from numba_cuda_mlir.descriptor import MLIRTargetContext

registry = LoweringRegistry()
lower = registry.lower
from numba_cuda_mlir.logging import trace
from numba_cuda_mlir.lowering_utilities import (
    index_of,
    convert,
    is_complex_type as _is_complex_type,
    get_llvm_struct_for_complex as _get_llvm_struct_for_complex,
    complex_to_llvm_struct as _complex_to_llvm_struct,
    llvm_struct_to_complex as _llvm_struct_to_complex,
)


def get_record_field_ptr(builder, record_ptr, record_type, field_name):
    """
    Get a pointer to a field within a record.

    Args:
        builder: MLIRLower builder
        record_ptr: LLVM pointer to the record's byte storage
        record_type: The Numba Record type
        field_name: Name of the field to access

    Returns:
        LLVM pointer to the field
    """
    offset = record_type.offset(field_name)
    field_numba_type = record_type.typeof(field_name)

    trace(f"get_record_field_ptr: field={field_name}, offset={offset}, type={field_numba_type}")

    # GEP with byte offset
    # llvm.getelementptr ptr[offset] : (!llvm.ptr) -> !llvm.ptr
    offset_val = arith_dialect.constant(T.i64(), offset)
    field_ptr = llvm.getelementptr(
        llvm.PointerType.get(),
        record_ptr,
        [offset_val],
        [LLVM_DYNAMIC],
        T.i8(),
        None,
    )

    return field_ptr, field_numba_type


@registry.lower_getattr_generic(Record)
def lower_record_getattr(
    context: MLIRTargetContext,
    builder: "MLIRLower",
    target: numba_ir.Var,
    value: numba_ir.Var,
    attr: str,
):
    """
    Lower getattr for Record field access.

    record.field_name -> load value from record at field offset
    """
    record_type = builder.get_numba_type(value.name)
    trace("Lowering Record getattr: %s.%s", record_type, attr)

    # Check field exists
    if attr not in record_type.fields:
        available = list(record_type.fields.keys())
        raise AttributeError(f"Record has no field '{attr}'. Available fields: {available}")

    # Load the record pointer
    record_ptr = builder.load_var(value)

    # Get pointer to the field
    field_ptr, field_numba_type = get_record_field_ptr(builder, record_ptr, record_type, attr)

    if isinstance(field_numba_type, NestedArray):
        # Nested array field - create a memref pointing to the data
        result = _load_nested_array_field(builder, field_ptr, field_numba_type, record_type)
    elif isinstance(field_numba_type, Record):
        result = field_ptr
    else:
        # Scalar field - load directly
        field_mlir_type = builder.get_mlir_type(field_numba_type)
        field_storage_type = builder.get_storage_type(field_numba_type)
        if _is_complex_type(field_mlir_type):
            # Complex types: load as LLVM struct, then convert to complex
            struct_type = _get_llvm_struct_for_complex(field_mlir_type)
            struct_val = llvm.load(struct_type, field_ptr)
            result = _llvm_struct_to_complex(struct_val, field_mlir_type)
        else:
            stored = llvm.load(field_storage_type, field_ptr)
            result = builder.from_storage(field_numba_type, stored)

    builder.store_var(target, result)
    trace("Stored Record field '%s' into %s", attr, target.name)


def _load_nested_array_field(builder, field_ptr, nested_type, record_type):
    """
    Load a nested array field from a record.

    Returns the LLVM pointer to the nested array data directly.
    Shape and stride information is available from nested_type at compile time.
    """
    trace(
        "Loading nested array field: shape=%s, dtype=%s",
        nested_type.shape,
        nested_type.dtype,
    )

    # Just return the pointer to the field data directly
    # The caller will use nested_type for shape/stride info when indexing
    trace("Returning pointer to nested array data")
    return field_ptr


@registry.lower_setattr_generic(Record)
def lower_record_setattr(
    context: MLIRTargetContext,
    builder: "MLIRLower",
    sig,
    args,
    attr: str,
):
    """
    Lower setattr for Record field assignment.

    record.field_name = value -> store value to record at field offset
    """
    record_type, value_type = sig.args
    record_var, value_var = args

    trace("Lowering Record setattr: %s.%s = %s", record_type, attr, value_type)

    # Check field exists
    if attr not in record_type.fields:
        available = list(record_type.fields.keys())
        raise AttributeError(f"Record has no field '{attr}'. Available fields: {available}")

    # Load the record pointer and value
    record_ptr = builder.load_var(record_var)
    value = builder.load_var(value_var)

    # Get pointer to the field
    field_ptr, field_numba_type = get_record_field_ptr(builder, record_ptr, record_type, attr)

    if isinstance(field_numba_type, NestedArray):
        # Nested array field - need to copy data
        _store_nested_array_field(builder, field_ptr, value, field_numba_type, value_type)
    else:
        # Scalar field - store directly
        field_mlir_type = builder.get_mlir_type(field_numba_type)
        field_storage_type = builder.get_storage_type(field_numba_type)

        # Convert value to field type if needed
        value = convert(value, field_mlir_type)

        if _is_complex_type(field_mlir_type):
            # Complex types: convert to LLVM struct before storing
            value = _complex_to_llvm_struct(value)
        else:
            value = builder.as_storage(field_numba_type, value)

        llvm.store(value, field_ptr)

    trace("Stored value into Record field '%s'", attr)


def _store_nested_array_field(builder, field_ptr, value, nested_type, value_type):
    """
    Store a value into a nested array field.

    This copies data from the source array into the nested array storage.
    Source may be a memref or LLVM pointer, destination is an LLVM pointer.
    Uses scf.for to iterate over elements, supporting any rank.
    """
    trace("Storing into nested array field: %s", nested_type)

    elem_size = storage_itemsize_bytes(nested_type.dtype)
    elem_mlir_type = builder.get_mlir_type(nested_type.dtype)
    elem_storage_type = builder.get_storage_type(nested_type.dtype)

    # Determine if source is memref or LLVM pointer
    is_source_memref = isinstance(value.type, ir.MemRefType)
    is_source_llvm_ptr = isinstance(value.type, llvm.PointerType)

    trace(
        "Source type: %s, is_memref=%s, is_llvm_ptr=%s",
        value.type,
        is_source_memref,
        is_source_llvm_ptr,
    )

    shape = list(nested_type.shape)
    strides = list(nested_type.strides)
    ndim = len(shape)

    # Compute total number of elements
    total_elements = 1
    for s in shape:
        total_elements *= s

    # Use scf.for to iterate over linear index, then decompose to multi-dim indices
    for linear_idx in scf.for_(0, total_elements, 1):
        # Convert linear index to i64 for arithmetic
        linear_i64 = arith.index_cast(T.i64(), linear_idx)

        # Decompose linear index to multi-dimensional indices (row-major order)
        # For shape [A, B, C]: idx0 = linear // (B*C), idx1 = (linear // C) % B, idx2 = linear % C
        indices = []
        remaining = linear_i64
        for dim in range(ndim):
            # Compute product of trailing dimensions
            trailing_product = 1
            for d in range(dim + 1, ndim):
                trailing_product *= shape[d]
            trailing_val = arith_dialect.constant(T.i64(), trailing_product)

            if trailing_product > 1:
                idx = arith.divui(remaining, trailing_val)
                remaining = arith.remui(remaining, trailing_val)
            else:
                idx = remaining
            indices.append(idx)

        # Load from source
        if is_source_memref:
            # Convert indices to index type for memref.load
            memref_indices = [arith.index_cast(T.index(), idx) for idx in indices]
            src_elem = memref_dialect.load(value, memref_indices)
            src_elem = builder.from_storage(nested_type.dtype, src_elem)
        elif is_source_llvm_ptr:
            # Linear byte offset for source (row-major, contiguous)
            src_byte_offset = arith.muli(linear_i64, arith_dialect.constant(T.i64(), elem_size))
            src_elem_ptr = llvm.getelementptr(
                llvm.PointerType.get(),
                value,
                [src_byte_offset],
                [LLVM_DYNAMIC],
                T.i8(),
                None,
            )
            src_stored = llvm.load(elem_storage_type, src_elem_ptr)
            src_elem = builder.from_storage(nested_type.dtype, src_stored)
        else:
            raise NotImplementedError(
                f"Cannot store from source type {value.type} into nested array"
            )

        # Destination byte offset using strides: sum(idx[i] * stride[i])
        dst_byte_offset = arith_dialect.constant(T.i64(), 0)
        for dim in range(ndim):
            stride_val = arith_dialect.constant(T.i64(), strides[dim])
            term = arith.muli(indices[dim], stride_val)
            dst_byte_offset = arith.addi(dst_byte_offset, term)

        dst_elem_ptr = llvm.getelementptr(
            llvm.PointerType.get(),
            field_ptr,
            [dst_byte_offset],
            [LLVM_DYNAMIC],
            T.i8(),
            None,
        )

        # Handle complex types
        if _is_complex_type(elem_mlir_type):
            src_elem = _complex_to_llvm_struct(src_elem)
        else:
            src_elem = builder.as_storage(nested_type.dtype, src_elem)

        llvm.store(src_elem, dst_elem_ptr)
        scf.yield_([])


# Static getitem/setitem for bracket notation: record['field']


@lower("static_getitem", Record, types.StringLiteral)
@lower(operator.getitem, Record, types.StringLiteral)
def lower_record_static_getitem_str(builder, target, args, kwargs):
    """
    Lower record['field_name'] to getattr.
    """
    record_var = args[0]
    field_name = args[1]  # This is the literal string value

    record_type = builder.get_numba_type(record_var.name)

    # Get the field name from the string literal
    field_name_type = builder.get_numba_type(field_name)
    attr = field_name_type.literal_value

    trace("Lowering Record static_getitem: %s['%s']", record_type, attr)

    # Delegate to getattr lowering
    # We need to manually invoke the getattr logic
    record_ptr = builder.load_var(record_var)

    if attr not in record_type.fields:
        available = list(record_type.fields.keys())
        raise AttributeError(f"Record has no field '{attr}'. Available fields: {available}")

    field_ptr, field_numba_type = get_record_field_ptr(builder, record_ptr, record_type, attr)

    if isinstance(field_numba_type, NestedArray):
        result = _load_nested_array_field(builder, field_ptr, field_numba_type, record_type)
    elif isinstance(field_numba_type, Record):
        result = field_ptr
    else:
        field_mlir_type = builder.get_mlir_type(field_numba_type)
        field_storage_type = builder.get_storage_type(field_numba_type)
        if _is_complex_type(field_mlir_type):
            # Complex types: load as LLVM struct, then convert to complex
            struct_type = _get_llvm_struct_for_complex(field_mlir_type)
            struct_val = llvm.load(struct_type, field_ptr)
            result = _llvm_struct_to_complex(struct_val, field_mlir_type)
        else:
            stored = llvm.load(field_storage_type, field_ptr)
            result = builder.from_storage(field_numba_type, stored)

    builder.store_var(target, result)


@lower("static_getitem", Record, types.IntegerLiteral)
@lower(operator.getitem, Record, types.IntegerLiteral)
def lower_record_static_getitem_int(builder, target, args, kwargs):
    """
    Lower record[0] to getattr using field index.
    """
    record_var = args[0]
    index_var = args[1]

    record_type = builder.get_numba_type(record_var.name)

    # Get the field index from the integer literal
    index_type = builder.get_numba_type(index_var)
    field_index = index_type.literal_value

    # Get field name by index
    field_names = list(record_type.fields.keys())
    if field_index < 0 or field_index >= len(field_names):
        raise IndexError(
            f"Record field index {field_index} out of range (0-{len(field_names) - 1})"
        )
    attr = field_names[field_index]

    trace(
        "Lowering Record static_getitem: %s[%s] -> field '%s'",
        record_type,
        field_index,
        attr,
    )

    record_ptr = builder.load_var(record_var)

    field_ptr, field_numba_type = get_record_field_ptr(builder, record_ptr, record_type, attr)

    if isinstance(field_numba_type, NestedArray):
        result = _load_nested_array_field(builder, field_ptr, field_numba_type, record_type)
    elif isinstance(field_numba_type, Record):
        result = field_ptr
    else:
        field_mlir_type = builder.get_mlir_type(field_numba_type)
        field_storage_type = builder.get_storage_type(field_numba_type)
        if _is_complex_type(field_mlir_type):
            struct_type = _get_llvm_struct_for_complex(field_mlir_type)
            struct_val = llvm.load(struct_type, field_ptr)
            result = _llvm_struct_to_complex(struct_val, field_mlir_type)
        else:
            stored = llvm.load(field_storage_type, field_ptr)
            result = builder.from_storage(field_numba_type, stored)

    builder.store_var(target, result)


@lower("static_setitem", Record, types.StringLiteral, types.Any)
@lower(operator.setitem, Record, types.StringLiteral, types.Any)
def lower_record_static_setitem_str(builder, target, args, kwargs):
    """
    Lower record['field_name'] = value to setattr.
    """
    record_var = args[0]
    field_name_var = args[1]
    value_var = args[2]

    record_type = builder.get_numba_type(record_var.name)
    value_type = builder.get_numba_type(value_var.name)

    # Get the field name from the string literal
    field_name_type = builder.get_numba_type(field_name_var)
    attr = field_name_type.literal_value

    trace("Lowering Record static_setitem: %s['%s'] = %s", record_type, attr, value_type)

    record_ptr = builder.load_var(record_var)
    value = builder.load_var(value_var)

    if attr not in record_type.fields:
        available = list(record_type.fields.keys())
        raise AttributeError(f"Record has no field '{attr}'. Available fields: {available}")

    field_ptr, field_numba_type = get_record_field_ptr(builder, record_ptr, record_type, attr)

    if isinstance(field_numba_type, NestedArray):
        _store_nested_array_field(builder, field_ptr, value, field_numba_type, value_type)
    else:
        field_mlir_type = builder.get_mlir_type(field_numba_type)
        field_storage_type = builder.get_storage_type(field_numba_type)
        value = convert(value, field_mlir_type)
        if _is_complex_type(field_mlir_type):
            # Complex types: convert to LLVM struct before storing
            value = _complex_to_llvm_struct(value)
        else:
            value = builder.as_storage(field_numba_type, value)
        llvm.store(value, field_ptr)


# NestedArray getitem/setitem - work with raw LLVM pointers


def _compute_nested_array_elem_ptr(builder, base_ptr, nested_type, indices):
    """
    Compute pointer to element in nested array using LLVM pointer arithmetic.

    Args:
        base_ptr: LLVM pointer to start of nested array data
        nested_type: The NestedArray type (contains shape, strides, dtype)
        indices: List of index values (LLVM i64)

    Returns:
        LLVM pointer to the element
    """
    elem_size = nested_type.dtype.bitwidth // 8
    strides = list(nested_type.strides)

    # Compute linear byte offset: sum(index[i] * stride[i])
    byte_offset = None
    for idx, stride in zip(indices, strides):
        # stride is in bytes
        stride_val = arith_dialect.constant(T.i64(), stride)
        term = arith.muli(idx, stride_val)
        if byte_offset is None:
            byte_offset = term
        else:
            byte_offset = arith.addi(byte_offset, term)

    if byte_offset is None:
        # Scalar access (0-D nested array?)
        byte_offset = arith_dialect.constant(T.i64(), 0)

    # GEP to compute element pointer
    elem_ptr = llvm.getelementptr(
        llvm.PointerType.get(),
        base_ptr,
        [byte_offset],
        [LLVM_DYNAMIC],
        T.i8(),
        None,
    )

    return elem_ptr


def lower_nested_array_getitem_int(builder, target, args, kwargs):
    """
    Lower nested_array[i] for 1D nested arrays.
    """
    array_var = args[0]
    index_var = args[1]

    nested_type = builder.get_numba_type(array_var.name)
    trace("Lowering NestedArray getitem: %s[int]", nested_type)

    # Load the base pointer and index
    base_ptr = builder.load_var(array_var)
    index = builder.load_var(index_var)

    # Convert index to i64
    index = convert(index, T.i64())

    # Compute element pointer
    elem_ptr = _compute_nested_array_elem_ptr(builder, base_ptr, nested_type, [index])

    # Load and return the element
    elem_mlir_type = builder.get_mlir_type(nested_type.dtype)
    if isinstance(nested_type.dtype, Record):
        result = elem_ptr
    elif _is_complex_type(elem_mlir_type):
        struct_type = _get_llvm_struct_for_complex(elem_mlir_type)
        struct_val = llvm.load(struct_type, elem_ptr)
        result = _llvm_struct_to_complex(struct_val, elem_mlir_type)
    else:
        result = llvm.load(elem_mlir_type, elem_ptr)

    builder.store_var(target, result)
    trace("Loaded NestedArray element into %s", target.name)


def lower_nested_array_setitem_int(builder, target, args, kwargs):
    """
    Lower nested_array[i] = value for 1D nested arrays.
    """
    array_var = args[0]
    index_var = args[1]
    value_var = args[2]

    nested_type = builder.get_numba_type(array_var.name)
    trace("Lowering NestedArray setitem: %s[int] = value", nested_type)

    # Load the base pointer, index, and value
    base_ptr = builder.load_var(array_var)
    index = builder.load_var(index_var)
    value = builder.load_var(value_var)

    # Convert index to i64
    index = convert(index, T.i64())

    # Compute element pointer
    elem_ptr = _compute_nested_array_elem_ptr(builder, base_ptr, nested_type, [index])

    # Convert value to element type if needed
    elem_mlir_type = builder.get_mlir_type(nested_type.dtype)
    value = convert(value, elem_mlir_type)

    # Store the value
    if _is_complex_type(elem_mlir_type):
        value = _complex_to_llvm_struct(value)
    llvm.store(value, elem_ptr)

    trace("Stored value into NestedArray element")


def _unpack_tuple_indices(indices_val, num_indices):
    """Unpack tuple indices from a Python tuple or LLVM struct."""
    indices = []
    if isinstance(indices_val, tuple):
        indices.extend(convert(idx, T.i64()) for idx in indices_val)
    elif hasattr(indices_val, "type"):
        for i in range(num_indices):
            idx = llvm.extractvalue(
                T.i64(),
                indices_val,
                position=ir.DenseI64ArrayAttr.get([i]),
            )
            indices.append(idx)
    else:
        raise NotImplementedError(f"Cannot unpack indices of type {type(indices_val)}")
    return indices


def lower_nested_array_getitem_tuple(builder, target, args, kwargs):
    """
    Lower nested_array[i, j, ...] for multi-dimensional nested arrays.
    """
    array_var = args[0]
    indices_var = args[1]

    nested_type = builder.get_numba_type(array_var.name)
    indices_type = builder.get_numba_type(indices_var.name)

    trace("Lowering NestedArray getitem: %s[tuple]", nested_type)

    base_ptr = builder.load_var(array_var)
    indices_val = builder.load_var(indices_var)
    indices = _unpack_tuple_indices(indices_val, len(indices_type))

    elem_ptr = _compute_nested_array_elem_ptr(builder, base_ptr, nested_type, indices)

    # Load and return the element
    elem_mlir_type = builder.get_mlir_type(nested_type.dtype)
    if isinstance(nested_type.dtype, Record):
        result = elem_ptr
    elif _is_complex_type(elem_mlir_type):
        struct_type = _get_llvm_struct_for_complex(elem_mlir_type)
        struct_val = llvm.load(struct_type, elem_ptr)
        result = _llvm_struct_to_complex(struct_val, elem_mlir_type)
    else:
        result = llvm.load(elem_mlir_type, elem_ptr)

    builder.store_var(target, result)
    trace("Loaded NestedArray element into %s", target.name)


def lower_nested_array_setitem_tuple(builder, target, args, kwargs):
    """
    Lower nested_array[i, j, ...] = value for multi-dimensional nested arrays.
    """
    array_var = args[0]
    indices_var = args[1]
    value_var = args[2]

    nested_type = builder.get_numba_type(array_var.name)
    indices_type = builder.get_numba_type(indices_var.name)

    trace("Lowering NestedArray setitem: %s[tuple] = value", nested_type)

    base_ptr = builder.load_var(array_var)
    value = builder.load_var(value_var)
    indices_val = builder.load_var(indices_var)
    indices = _unpack_tuple_indices(indices_val, len(indices_type))

    elem_ptr = _compute_nested_array_elem_ptr(builder, base_ptr, nested_type, indices)

    # Convert value to element type and store
    elem_mlir_type = builder.get_mlir_type(nested_type.dtype)
    value = convert(value, elem_mlir_type)

    if _is_complex_type(elem_mlir_type):
        value = _complex_to_llvm_struct(value)

    llvm.store(value, elem_ptr)
    trace("Stored value into NestedArray element")



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
import functools
from numba_cuda_mlir._mlir.dialects import gpu, arith, llvm
from numba_cuda_mlir._mlir import ir


def recursively_apply(pattern):
    @functools.wraps(pattern)
    def wrapper(op):
        if pattern(op):
            return True
        for region in op.regions:
            for block in region.blocks:
                for operation in block:
                    if recursively_apply(pattern)(operation):
                        return True
        return False

    return wrapper


@recursively_apply
def fixup_nvvm_arg_attrs(op: gpu.GPUFuncOp):
    if "numba_cuda_mlir.arg_attrs" in op.attributes:
        attrs = [[j for j in i] for i in op.attributes["numba_cuda_mlir.arg_attrs"]]
        if all(len(i) == 0 for i in attrs):
            return
        orig_arg_types = op.attributes["numba_cuda_mlir.orig_arg_types"]
        arg_attrs = op.attributes["numba_cuda_mlir.arg_attrs"]
        new_arg_attrs = []
        for i, arg_attr in enumerate(arg_attrs):
            new_arg_attr = {}
            for namedattr in arg_attr:
                if "numba_cuda_mlir.grid_constant" == namedattr.name:
                    new_arg_attr["nvvm.grid_constant"] = ir.UnitAttr.get()
                else:
                    new_arg_attr[namedattr.name] = namedattr.attr

            new_arg_attrs.append(ir.DictAttr.get(new_arg_attr))
            orig_arg_type = orig_arg_types[i].value

            # assuming the CAPI for memref types is expanded here
            if isinstance(orig_arg_type, ir.MemRefType):
                # append again for the aligned pointer
                new_arg_attrs.append(ir.DictAttr.get(new_arg_attr))

                # Empty attr for the offset
                new_arg_attrs.append(ir.DictAttr.get({}))

                # Give empty attrs for the shape args
                r = orig_arg_type.rank
                new_arg_attrs.extend([ir.DictAttr.get({}) for _ in range(r)])

                # Give empty attrs for the stride args
                new_arg_attrs.extend([ir.DictAttr.get({}) for _ in range(r)])

        op.attributes["arg_attrs"] = ir.ArrayAttr.get(new_arg_attrs)
        del op.attributes["numba_cuda_mlir.arg_attrs"]
        del op.attributes["numba_cuda_mlir.orig_arg_types"]


_EXOTIC_FLOAT_TYPES = frozenset(
    [
        "f4E2M1FN",
        "f6E2M3FN",
        "f6E3M2FN",
        "f8E3M4",
        "f8E4M3B11FNUZ",
        "f8E4M3FN",
        "f8E4M3FNUZ",
        "f8E4M3",
        "f8E5M2FNUZ",
        "f8E5M2",
        "f8E8M0FNU",
        "tf32",
    ]
)


def _is_exotic_float(ty: ir.Type) -> bool:
    return isinstance(ty, ir.FloatType) and str(ty) in _EXOTIC_FLOAT_TYPES


def _resolve_exotic_float_casts(module: ir.Module):
    """Replace unrealized_conversion_cast between integer and exotic float
    types with arith.bitcast. MLIR's memref-to-LLVM lowering inserts these
    casts for sub-32-bit float element types because LLVM has no native
    representation for them."""
    worklist = []

    def collect(op):
        if op.operation.name == "builtin.unrealized_conversion_cast":
            if len(op.results) == 1 and len(op.operands) == 1:
                src_ty = op.operands[0].type
                dst_ty = op.results[0].type
                if (isinstance(src_ty, ir.IntegerType) and _is_exotic_float(dst_ty)) or (
                    _is_exotic_float(src_ty) and isinstance(dst_ty, ir.IntegerType)
                ):
                    worklist.append(op)
        for region in op.operation.regions:
            for block in region.blocks:
                for child in block:
                    collect(child)

    collect(module.operation.opview)

    for op in worklist:
        src = op.operands[0]
        dst_ty = op.results[0].type
        loc = op.operation.location
        with ir.InsertionPoint(op), loc:
            bc = arith.bitcast(dst_ty, src)
        op.results[0].replace_all_uses_with(bc)
        op.operation.erase()


_SHARED_ADDRESS_SPACE = 3


def _is_shared_llvm_ptr(ty: ir.Type) -> bool:
    return (
        isinstance(ty, llvm.PointerType)
        and llvm.PointerType(ty).address_space == _SHARED_ADDRESS_SPACE
    )


def _bit_storage_type_for_float(ty: ir.Type):
    if not isinstance(ty, ir.FloatType):
        return None
    from numba_cuda_mlir.models import get_float_integer_storage_map

    width = get_float_integer_storage_map().get(str(ty))
    if width is None:
        return None
    return ir.IntegerType.get_signless(width)


def _copy_op_attrs(src, dst):
    for name in src.operation.attributes:
        dst.operation.attributes[name] = src.operation.attributes[name]


def _resolve_shared_bit_storage_float_accesses(module: ir.Module):
    """Rewrite shared-memory LLVM scalar loads/stores so float operands use integer storage.

    For MLIR floating-point types whose ABI/storage representation is wider integer bits
    (half/bfloat16, eight-bit storages for sub-byte floats, TF32), load/store through the
    integer representation when the pointer is shared (address space 3).

    nvjitlink LTO can drop certain half-precision scalar stores before a widened load; forcing
    integer accesses preserves bit patterns across that optimization.
    """
    worklist = []

    def collect(op):
        if op.operation.name in ("llvm.load", "llvm.store"):
            worklist.append(op)
        for region in op.operation.regions:
            for block in region.blocks:
                for child in block:
                    collect(child)

    collect(module.operation.opview)

    for op in worklist:
        if op.operation.name == "llvm.store":
            value = op.operands[0]
            addr = op.operands[1]
            storage_type = _bit_storage_type_for_float(value.type)
            if storage_type is None or not _is_shared_llvm_ptr(addr.type):
                continue
            loc = op.operation.location
            with ir.InsertionPoint(op), loc:
                bits = llvm.bitcast(storage_type, value)
                new_store = llvm.store(bits, addr)
            _copy_op_attrs(op, new_store)
            op.operation.erase()
            continue

        result = op.results[0]
        addr = op.operands[0]
        storage_type = _bit_storage_type_for_float(result.type)
        if storage_type is None or not _is_shared_llvm_ptr(addr.type):
            continue
        loc = op.operation.location
        with ir.InsertionPoint(op), loc:
            bits = llvm.load(storage_type, addr)
            value = llvm.bitcast(result.type, bits)
        _copy_op_attrs(op, bits.owner.opview)
        result.replace_all_uses_with(value)
        op.operation.erase()


def run_pre_codegen_patterns(module: ir.Module):
    fixup_nvvm_arg_attrs(module.operation)
    _resolve_exotic_float_casts(module)
    _resolve_shared_bit_storage_float_accesses(module)
    # TODO(ajm): why does this not trigger?
    # patterns = RewritePatternSet()
    # patterns.add(gpu.GPUFuncOp, fixup_nvvm_arg_attrs)
    # frozen = patterns.freeze()
    # apply_patterns_and_fold_greedily(module, frozen)



#  Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
#  See https://llvm.org/LICENSE.txt for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from ._arith_ops_gen import *
from ._arith_ops_gen import _Dialect
from ._arith_enum_gen import *
from array import array as _array
from typing import overload

try:
    from ..ir import *
    from ._ods_common import (
        get_default_loc_context as _get_default_loc_context,
        _cext as _ods_cext,
        get_op_result_or_op_results as _get_op_result_or_op_results,
    )

    from typing import Any, List, Union
except ImportError as e:
    raise RuntimeError("Error loading imports from extension module") from e


@_ods_cext.register_operation(_Dialect, replace=True)
class ConstantOp(ConstantOp):
    """Specialization for the constant op class."""

    @overload
    def __init__(self, value: Attribute, *, loc=None, ip=None):
        ...

    @overload
    def __init__(
        self, result: Type, value: Union[int, float, _array], *, loc=None, ip=None
    ):
        ...

    def __init__(self, result, value, *, loc=None, ip=None):
        if value is None:
            assert isinstance(result, Attribute)
            super().__init__(result, loc=loc, ip=ip)
            return

        if isinstance(value, int):
            super().__init__(IntegerAttr.get(result, value), loc=loc, ip=ip)
        elif isinstance(value, float):
            super().__init__(FloatAttr.get(result, value), loc=loc, ip=ip)
        elif isinstance(value, _array):
            if 8 * value.itemsize != result.element_type.width:
                raise ValueError(
                    f"Mismatching array element ({8 * value.itemsize}) and type ({result.element_type.width}) width."
                )
            if value.typecode in ["i", "l", "q"]:
                super().__init__(DenseIntElementsAttr.get(value, type=result))
            elif value.typecode in ["f", "d"]:
                super().__init__(DenseFPElementsAttr.get(value, type=result))
            else:
                raise ValueError(f'Unsupported typecode: "{value.typecode}".')
        else:
            super().__init__(value, loc=loc, ip=ip)

    @classmethod
    def create_index(cls, value: int, *, loc=None, ip=None):
        """Create an index-typed constant."""
        return cls(
            IndexType.get(context=_get_default_loc_context(loc)), value, loc=loc, ip=ip
        )

    @property
    def type(self):
        return self.results[0].type

    @property
    def value(self):
        return self.operation.attributes["value"]

    @property
    def literal_value(self) -> Union[int, float]:
        if isinstance(self.type, (IntegerType, IndexType)):
            return IntegerAttr(self.value).value
        elif isinstance(self.type, FloatType):
            return FloatAttr(self.value).value
        else:
            raise ValueError("only integer and float constants have literal values")


def constant(
    result: Type, value: Union[int, float, Attribute, _array], *, loc=None, ip=None
) -> Value:
    return _get_op_result_or_op_results(ConstantOp(result, value, loc=loc, ip=ip))



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# local_array_from transformation pass
# Converts local_array_from(genexp, dtype) to local_array + loop
import ast

from numba_cuda_mlir.ast_transforms.pipeline import ASTTransformPass, TransformContext


class LocalArrayFromTransformer(ast.NodeTransformer):
    """AST transformer that converts local_array_from calls to local_array + loop.

    Transforms:
        arr = cuda.local_array_from((expr for var in iterable), dtype=np.float32)

    Into:
        arr = cuda.local_array(len(iterable), dtype=np.float32)
        for __i, var in enumerate(iterable):
            arr[__i] = expr
    """

    def __init__(self):
        self.modified = False
        self._counter = 0

    def _make_temp_name(self) -> str:
        name = f"__laf_i_{self._counter}"
        self._counter += 1
        return name

    def _is_local_array_from(self, node: ast.expr) -> bool:
        """Check if node is a call to local_array_from."""
        if not isinstance(node, ast.Call):
            return False
        func = node.func
        # Match: local_array_from(...) or X.local_array_from(...)
        if isinstance(func, ast.Name) and func.id == "local_array_from":
            return True
        if isinstance(func, ast.Attribute) and func.attr == "local_array_from":
            return True
        return False

    def _get_local_array_func(self, original_func: ast.expr) -> ast.expr:
        """Convert local_array_from to cuda.local_array.

        Always produces ``cuda.local_array`` so callers can reference
        ``local_array_from`` from ``cuda.experimental``.
        """
        return ast.Attribute(
            value=ast.Name(id="cuda", ctx=ast.Load()),
            attr="local_array",
            ctx=ast.Load(),
        )

    def visit_Assign(self, node: ast.Assign) -> ast.AST | list[ast.stmt]:
        node = self.generic_visit(node)

        if len(node.targets) != 1 or not self._is_local_array_from(node.value):
            return node

        call = node.value
        if not call.args or not isinstance(call.args[0], ast.GeneratorExp):
            return node

        genexp = call.args[0]
        if len(genexp.generators) != 1:
            # Only handle single-loop generators for now
            return node

        gen = genexp.generators[0]
        if gen.ifs:
            # Don't handle conditional generators for now
            return node

        self.modified = True
        arr_name = node.targets[0]
        if not isinstance(arr_name, ast.Name):
            return node
        arr_name_str = arr_name.id
        idx_name = self._make_temp_name()

        # Build: arr = cuda.local_array(len(iterable), dtype=dtype)
        len_call = ast.Call(
            func=ast.Name(id="len", ctx=ast.Load()),
            args=[gen.iter],
            keywords=[],
        )
        local_array_call = ast.Call(
            func=self._get_local_array_func(call.func),
            args=[len_call],
            keywords=call.keywords,  # Pass through dtype and other kwargs
        )
        arr_assign = ast.Assign(
            targets=[ast.Name(id=arr_name_str, ctx=ast.Store())],
            value=local_array_call,
        )

        # Build: for idx in range(len(iterable)):
        #            var = iterable[idx]
        #            arr[idx] = expr
        var_assign = ast.Assign(
            targets=[gen.target],
            value=ast.Subscript(
                value=gen.iter,
                slice=ast.Name(id=idx_name, ctx=ast.Load()),
                ctx=ast.Load(),
            ),
        )
        arr_store = ast.Assign(
            targets=[
                ast.Subscript(
                    value=ast.Name(id=arr_name_str, ctx=ast.Load()),
                    slice=ast.Name(id=idx_name, ctx=ast.Load()),
                    ctx=ast.Store(),
                )
            ],
            value=genexp.elt,
        )
        for_loop = ast.For(
            target=ast.Name(id=idx_name, ctx=ast.Store()),
            iter=ast.Call(
                func=ast.Name(id="range", ctx=ast.Load()),
                args=[len_call],
                keywords=[],
            ),
            body=[var_assign, arr_store],
            orelse=[],
        )

        return [arr_assign, for_loop]


class StatementExpander(ast.NodeTransformer):
    """Expand statement lists that contain lists from transforms."""

    def _expand_body(self, stmts: list[ast.stmt]) -> list[ast.stmt]:
        result = []
        for stmt in stmts:
            visited = self.visit(stmt)
            if isinstance(visited, list):
                result.extend(visited)
            else:
                result.append(visited)
        return result

    def visit_FunctionDef(self, node: ast.FunctionDef) -> ast.FunctionDef:
        node.body = self._expand_body(node.body)
        return node

    def visit_If(self, node: ast.If) -> ast.If:
        node.body = self._expand_body(node.body)
        node.orelse = self._expand_body(node.orelse)
        return node

    def visit_For(self, node: ast.For) -> ast.For:
        node.body = self._expand_body(node.body)
        node.orelse = self._expand_body(node.orelse)
        return node

    def visit_While(self, node: ast.While) -> ast.While:
        node.body = self._expand_body(node.body)
        node.orelse = self._expand_body(node.orelse)
        return node


def transform_comprehensions(tree: ast.Module) -> tuple[ast.Module, bool]:
    """Transform local_array_from calls to local_array + loop.

    Returns (transformed_tree, was_modified).
    """
    transformer = LocalArrayFromTransformer()
    new_tree = transformer.visit(tree)
    if transformer.modified:
        expander = StatementExpander()
        new_tree = expander.visit(new_tree)
    ast.fix_missing_locations(new_tree)
    return new_tree, transformer.modified


class ComprehensionPass(ASTTransformPass):
    """Pipeline pass that transforms local_array_from calls to local_array + loop."""

    @property
    def name(self) -> str:
        return "LocalArrayFrom"

    def transform(self, tree: ast.Module, context: TransformContext) -> tuple[ast.Module, bool]:
        return transform_comprehensions(tree)



#  Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
#  See https://llvm.org/LICENSE.txt for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
"""YAML serialization is routed through here to centralize common logic."""

import sys


def multiline_str_representer(dumper, data):
    if len(data.splitlines()) > 1:
        return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")
    else:
        return dumper.represent_scalar("tag:yaml.org,2002:str", data)


try:
    from yaml import YAMLObject as _YAMLObject, add_representer

    add_representer(str, multiline_str_representer)
except ModuleNotFoundError as e:

    class _YAMLObject:
        pass


class YAMLObject(_YAMLObject):
    @classmethod
    def to_yaml(cls, dumper, self):
        """Default to a custom dictionary mapping."""
        return dumper.represent_mapping(cls.yaml_tag, self.to_yaml_custom_dict())

    def to_yaml_custom_dict(self):
        raise NotImplementedError()

    def as_linalg_yaml(self):
        return yaml_dump(self)


def yaml_dump(data, sort_keys=False, **kwargs):
    try:
        import yaml

        return yaml.dump(data, sort_keys=sort_keys, **kwargs)
    except ModuleNotFoundError as e:
        raise ModuleNotFoundError(
            f"This tool requires PyYAML but it was not installed. "
            f"Recommend: {sys.executable} -m pip install PyYAML"
        ) from e


def yaml_dump_all(data, sort_keys=False, explicit_start=True, **kwargs):
    try:
        import yaml

        return yaml.dump_all(
            data, sort_keys=sort_keys, explicit_start=explicit_start, **kwargs
        )
    except ModuleNotFoundError as e:
        raise ModuleNotFoundError(
            f"This tool requires PyYAML but it was not installed. "
            f"Recommend: {sys.executable} -m pip install PyYAML"
        ) from e


__all__ = [
    "yaml_dump",
    "yaml_dump_all",
    "YAMLObject",
]



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Protocol, Any, Tuple, runtime_checkable
from numba_cuda_mlir.numba_cuda import types


@runtime_checkable
class ArgumentHandler(Protocol):
    def prepare_args(
        self, ty: types.Type, val: Any, stream: Any = None, retr: list | None = None
    ) -> Tuple[types.Type, Any]: ...



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from numba_cuda_mlir.numba_cuda.types import Type, CPointer


class ByValPointerType(CPointer):
    def __init__(self, dtype: Type):
        super().__init__(dtype)
        self.__numba_cuda_mlir_attributes__ = {"llvm.byval": dtype}




# Autogenerated by mlir-tblgen; don't manually edit.

from ._ods_common import _cext as _ods_cext
from ._ods_common import (
    equally_sized_accessor as _ods_equally_sized_accessor,
    get_default_loc_context as _ods_get_default_loc_context,
    get_op_results_or_values as _get_op_results_or_values,
    segmented_accessor as _ods_segmented_accessor,
)
_ods_ir = _ods_cext.ir
_ods_cext.globals.register_traceback_file_exclusion(__file__)

import builtins
from typing import Any as _Any, Sequence as _Sequence, Union as _Union, Optional as _Optional
import sys as _sys
if _sys.version_info >= (3, 12):
  from collections.abc import Buffer as _Buffer  # pytype: disable=not-supported-yet
else:
  try:
    from typing_extensions import Buffer as _Buffer
  except ImportError:
    _Buffer = _Any


@_ods_cext.register_dialect
class _Dialect(_ods_ir.Dialect):
  DIALECT_NAMESPACE = "ub"

@_ods_cext.register_operation(_Dialect)
class PoisonOp(_ods_ir.OpView):
  r"""
  The `poison` operation materializes a compile-time poisoned constant value
  to indicate deferred undefined behavior.
  `value` attribute is needed to indicate an optional additional poison
  semantics (e.g. partially poisoned vectors), default value indicates results
  is fully poisoned.
  
  Examples:
  
  ```
  // Short form
  %0 = ub.poison : i32
  // Long form
  %1 = ub.poison <#custom_poison_elements_attr> : vector<4xi64>
  ```
  """

  OPERATION_NAME = "ub.poison"

  _ODS_REGIONS = (0, True)

  def __init__(self, result: _ods_ir.Type, *, value: _Optional[_Union[_Any, _ods_ir.Attribute]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    if value is not None: attributes["value"] = (value if (
        isinstance(value, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('PoisonAttrInterface')) else
          _ods_ir.AttrBuilder.get('PoisonAttrInterface')(value, context=_ods_context))
    results = []
    results.append(result)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def value(self) -> _ods_ir.Attribute:
    return self.operation.attributes["value"]

  @value.setter
  def value(self, value: _ods_ir.Attribute):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["value"] = value

  @builtins.property
  def result(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(PoisonOp)
class PoisonOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "ub.poison"

  @builtins.property
  def value(self) -> _ods_ir.Attribute:
    return self.attributes["value"]

def poison(result: _ods_ir.Type, *, value: _Optional[_Union[_Any, _ods_ir.Attribute]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return PoisonOp(result=result, value=value, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class UnreachableOp(_ods_ir.OpView):
  r"""
  The `unreachable` operation triggers immediate undefined behavior if
  executed.
  
  Example:
  
  ```
  ub.unreachable
  ```
  """

  OPERATION_NAME = "ub.unreachable"

  _ODS_REGIONS = (0, True)

  def __init__(self, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

@_ods_cext.register_op_adaptor(UnreachableOp)
class UnreachableOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "ub.unreachable"

def unreachable(*, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> UnreachableOp:
  return UnreachableOp(loc=loc, ip=ip)



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from numba_cuda_mlir.numba_cuda.typing import typeof
from numba_cuda_mlir.numba_cuda import types
from numba_cuda_mlir.numba_cuda.np import numpy_support
import numpy as np
import cupy


@typeof.typeof_impl.register(cupy.ndarray)
def typeof_cupy_ndarray(val, c):
    numba_dtype = numpy_support.from_dtype(np.dtype(val.dtype))
    return types.Array(dtype=numba_dtype, ndim=len(val.shape), layout="C")
